library angular2.src.debug.debug_element;

import "package:angular2/src/facade/lang.dart"
    show Type, isPresent, BaseException, isBlank;
import "package:angular2/src/facade/collection.dart"
    show List, ListWrapper, MapWrapper;
import "package:angular2/src/dom/dom_adapter.dart" show DOM;
import "package:angular2/src/core/compiler/element_injector.dart"
    show ElementInjector;
import "package:angular2/src/core/compiler/view.dart" show AppView;
import "package:angular2/src/core/compiler/view_ref.dart" show internalView;
import "package:angular2/src/core/compiler/element_ref.dart" show ElementRef;
import "package:angular2/src/render/dom/view/view.dart"
    show resolveInternalDomView;

/**
 * @exportedAs angular2/test
 *
 * An DebugElement contains information from the Angular compiler about an
 * element and provides access to the corresponding ElementInjector and
 * underlying dom Element, as well as a way to query for children.
 */
class DebugElement {
  AppView _parentView;
  num _boundElementIndex;
  ElementInjector _elementInjector;
  DebugElement(this._parentView, this._boundElementIndex) {
    this._elementInjector =
        this._parentView.elementInjectors[this._boundElementIndex];
  }
  static DebugElement create(ElementRef elementRef) {
    return new DebugElement(
        internalView(elementRef.parentView), elementRef.boundElementIndex);
  }
  dynamic get componentInstance {
    if (!isPresent(this._elementInjector)) {
      return null;
    }
    return this._elementInjector.getComponent();
  }
  dynamic get dynamicallyCreatedComponentInstance {
    if (!isPresent(this._elementInjector)) {
      return null;
    }
    return this._elementInjector.getDynamicallyLoadedComponent();
  }
  dynamic get domElement {
    return resolveInternalDomView(this._parentView.render).boundElements[
        this._boundElementIndex];
  }
  dynamic getDirectiveInstance(num directiveIndex) {
    return this._elementInjector.getDirectiveAtIndex(directiveIndex);
  }
  /**
   * Get child DebugElements from within the Light DOM.
   *
   * @return {List<DebugElement>}
   */
  List<DebugElement> get children {
    var thisElementBinder =
        this._parentView.proto.elementBinders[this._boundElementIndex];
    return this._getChildElements(this._parentView, thisElementBinder.index);
  }
  /**
   * Get the root DebugElement children of a component. Returns an empty
   * list if the current DebugElement is not a component root.
   *
   * @return {List<DebugElement>}
   */
  List<DebugElement> get componentViewChildren {
    var shadowView =
        this._parentView.componentChildViews[this._boundElementIndex];
    if (!isPresent(shadowView)) {
      // The current element is not a component.
      return ListWrapper.create();
    }
    return this._getChildElements(shadowView, null);
  }
  void triggerEventHandler(eventName, eventObj) {
    this._parentView.triggerEventHandlers(
        eventName, eventObj, this._boundElementIndex);
  }
  bool hasDirective(Type type) {
    if (!isPresent(this._elementInjector)) {
      return false;
    }
    return this._elementInjector.hasDirective(type);
  }
  dynamic inject(Type type) {
    if (!isPresent(this._elementInjector)) {
      return null;
    }
    return this._elementInjector.get(type);
  }
  /**
   * Return the first descendant TestElememt matching the given predicate
   * and scope.
   *
   * @param {Function: boolean} predicate
   * @param {Scope} scope
   *
   * @return {DebugElement}
   */
  DebugElement query(Function predicate, [scope = Scope.all]) {
    var results = this.queryAll(predicate, scope);
    return results.length > 0 ? results[0] : null;
  }
  /**
   * Return descendant TestElememts matching the given predicate
   * and scope.
   *
   * @param {Function: boolean} predicate
   * @param {Scope} scope
   *
   * @return {List<DebugElement>}
   */
  List<DebugElement> queryAll(Function predicate, [scope = Scope.all]) {
    var elementsInScope = scope(this);
    return ListWrapper.filter(elementsInScope, predicate);
  }
  List<DebugElement> _getChildElements(
      AppView view, num parentBoundElementIndex) {
    var els = ListWrapper.create();
    var parentElementBinder = null;
    if (isPresent(parentBoundElementIndex)) {
      parentElementBinder = view.proto.elementBinders[parentBoundElementIndex];
    }
    for (var i = 0; i < view.proto.elementBinders.length; ++i) {
      var binder = view.proto.elementBinders[i];
      if (binder.parent == parentElementBinder) {
        ListWrapper.push(els, new DebugElement(view, i));
        var views = view.viewContainers[i];
        if (isPresent(views)) {
          ListWrapper.forEach(views.views, (nextView) {
            els =
                ListWrapper.concat(els, this._getChildElements(nextView, null));
          });
        }
      }
    }
    return els;
  }
}
DebugElement inspectElement(ElementRef elementRef) {
  return DebugElement.create(elementRef);
}
/**
 * @exportedAs angular2/test
 */
class Scope {
  static List<DebugElement> all(debugElement) {
    var scope = ListWrapper.create();
    ListWrapper.push(scope, debugElement);
    ListWrapper.forEach(debugElement.children, (child) {
      scope = ListWrapper.concat(scope, Scope.all(child));
    });
    ListWrapper.forEach(debugElement.componentViewChildren, (child) {
      scope = ListWrapper.concat(scope, Scope.all(child));
    });
    return scope;
  }
  static List<DebugElement> light(debugElement) {
    var scope = ListWrapper.create();
    ListWrapper.forEach(debugElement.children, (child) {
      ListWrapper.push(scope, child);
      scope = ListWrapper.concat(scope, Scope.light(child));
    });
    return scope;
  }
  static List<DebugElement> view(debugElement) {
    var scope = ListWrapper.create();
    ListWrapper.forEach(debugElement.componentViewChildren, (child) {
      ListWrapper.push(scope, child);
      scope = ListWrapper.concat(scope, Scope.light(child));
    });
    return scope;
  }
}
/**
 * @exportedAs angular2/test
 */
class By {
  static Function all() {
    return (debugElement) => true;
  }
  static Function css(String selector) {
    return (debugElement) {
      return DOM.elementMatches(debugElement.domElement, selector);
    };
  }
  static Function directive(Type type) {
    return (debugElement) {
      return debugElement.hasDirective(type);
    };
  }
}
