library angular2.src.forms.directives.control_group_directive;

import "package:angular2/angular2.dart"
    show Directive, Ancestor, onDestroy, onInit;
import "package:angular2/di.dart" show Inject, Binding;
import "package:angular2/src/facade/collection.dart" show List, ListWrapper;
import "control_container_directive.dart" show ControlContainerDirective;
import "shared.dart" show controlPath;

const controlGroupBinding =
    const Binding(ControlContainerDirective, toAlias: ControlGroupDirective);
/**
 * Binds a ng-control group to a DOM element.
 *
 * # Example
 *
 * In this example, we create a ng-control group, and we bind the login and
 * password controls to the login and password elements.
 *
 * Here we use {@link formDirectives}, rather than importing each form directive individually, e.g.
 * `ControlDirective`, `ControlGroupDirective`. This is just a shorthand for the same end result.
 *
 *  ```
 * @Component({selector: "login-comp"})
 * @View({
 *      directives: [formDirectives],
 *      template:
 *              "<form [ng-form-model]='loginForm'>" +
 *              "<div ng-control-group="credentials">
 *              "Login <input type='text' ng-control='login'>" +
 *              "Password <input type='password' ng-control='password'>" +
 *              "<button (click)="onLogin()">Login</button>" +
 *              "</div>"
 *              "</form>"
 *      })
 * class LoginComp {
 *  loginForm:ControlGroup;
 *
 *  constructor() {
 *    this.loginForm = new ControlGroup({
 *      credentials: new ControlGroup({
 *        login: new Cntrol(""),
 *        password: new Control("")
 *      })
 *    });
 *  }
 *
 *  onLogin() {
 *    // this.loginForm.value
 *  }
 * }
 *
 *  ```
 *
 * @exportedAs angular2/forms
 */
@Directive(
    selector: "[ng-control-group]",
    hostInjector: const [controlGroupBinding],
    properties: const ["name: ng-control-group"],
    lifecycle: const [onInit, onDestroy])
class ControlGroupDirective extends ControlContainerDirective {
  ControlContainerDirective _parent;
  ControlGroupDirective(@Ancestor() ControlContainerDirective _parent)
      : super() {
    /* super call moved to initializer */;
    this._parent = _parent;
  }
  onInit() {
    this.formDirective.addControlGroup(this);
  }
  onDestroy() {
    this.formDirective.removeControlGroup(this);
  }
  List<String> get path {
    return controlPath(this.name, this._parent);
  }
  dynamic get formDirective {
    return this._parent.formDirective;
  }
}
