library angular2.src.forms.directives.control_name_directive;

import "package:angular2/src/facade/async.dart"
    show EventEmitter, ObservableWrapper;
import "package:angular2/src/facade/collection.dart"
    show List, StringMapWrapper, Map;
import "package:angular2/angular2.dart"
    show Directive, Ancestor, onDestroy, onChange;
import "package:angular2/di.dart" show Binding, Inject;
import "control_container_directive.dart" show ControlContainerDirective;
import "control_directive.dart" show ControlDirective;
import "shared.dart" show controlPath;
import "../model.dart" show Control;

const controlNameBinding =
    const Binding(ControlDirective, toAlias: ControlNameDirective);
/**
 * Binds a control with the specified name to a DOM element.
 *
 * # Example
 *
 * In this example, we bind the login control to an input element. When the value of the input
 * element
 * changes, the value of
 * the control will reflect that change. Likewise, if the value of the control changes, the input
 * element reflects that
 * change.
 *
 * Here we use {@link formDirectives}, rather than importing each form directive individually, e.g.
 * `ControlDirective`, `ControlGroupDirective`. This is just a shorthand for the same end result.
 *
 *  ```
 * @Component({selector: "login-comp"})
 * @View({
 *      directives: [formDirectives],
 *      template:
 *              "<form [ng-form-model]='loginForm'>" +
 *              "Login <input type='text' ng-control='login'>" +
 *              "<button (click)="onLogin()">Login</button>" +
 *              "</form>"
 *      })
 * class LoginComp {
 *  loginForm:ControlGroup;
 *
 *  constructor() {
 *    this.loginForm = new ControlGroup({
 *      login: new Control(""),
 *    });
 *  }
 *
 *  onLogin() {
 *    // this.loginForm.value
 *  }
 * }
 *
 *  ```
 *
 * @exportedAs angular2/forms
 */
@Directive(
    selector: "[ng-control]",
    hostInjector: const [controlNameBinding],
    properties: const ["name: ng-control", "model: ng-model"],
    events: const ["ngModel"],
    lifecycle: const [onDestroy, onChange])
class ControlNameDirective extends ControlDirective {
  ControlContainerDirective _parent;
  EventEmitter ngModel;
  dynamic model;
  bool _added;
  ControlNameDirective(@Ancestor() ControlContainerDirective _parent)
      : super() {
    /* super call moved to initializer */;
    this._parent = _parent;
    this.ngModel = new EventEmitter();
    this._added = false;
  }
  onChange(Map<String, dynamic> c) {
    if (!this._added) {
      this.formDirective.addControl(this);
      this._added = true;
    }
    if (StringMapWrapper.contains(c, "model")) {
      this.formDirective.updateModel(this, this.model);
    }
  }
  onDestroy() {
    this.formDirective.removeControl(this);
  }
  void viewToModelUpdate(dynamic newValue) {
    ObservableWrapper.callNext(this.ngModel, newValue);
  }
  List<String> get path {
    return controlPath(this.name, this._parent);
  }
  dynamic get formDirective {
    return this._parent.formDirective;
  }
  Control get control {
    return this.formDirective.getControl(this);
  }
}
